/*
   Copyright (c) 2014 Carsten Moeller, Pinneberg, Germany. <info@osm2po.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as 
   published by the Free Software Foundation, either version 3 of the 
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
   
   $Id: PgPolyWayWriter.java 4033 2015-12-22 09:44:31Z carsten $
*/

package de.cm.osm2po.plugins.postp;

import static de.cm.osm2po.misc.Utils.DF;
import static de.cm.osm2po.misc.Utils.bytesToHex;
import static de.cm.osm2po.misc.WkbUtils.toPolygonWkb;
import static de.cm.osm2po.primitives.VarString.toUTF8QuotedSqlBytes;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStream;
import java.util.Date;

import de.cm.osm2po.Config;
import de.cm.osm2po.Version;
import de.cm.osm2po.converter.JoinedWaysIterator;
import de.cm.osm2po.converter.PostProcessor;
import de.cm.osm2po.logging.Log;
import de.cm.osm2po.model.LatLons;
import de.cm.osm2po.model.Way;

public class PgPolyWayWriter implements PostProcessor {

    private static final byte KOMMA = ',';
    private static final byte SEMIKOLON = ';';
    private static final byte NEWLINE = '\n';
    
    @Override
    public void run(Config config, int index) throws Exception {
        
        File dir = config.getWorkDir();
        String prefix = config.getPrefix();
        Log log = config.getLog();
        
        String tableName = (prefix + "_2po_polyway").toLowerCase(); // Postgres-Problem
        String fileName = prefix + "_2po_polyway.sql";
        
        OutputStream os = null;
        File sqlOutFile = null;
        if (config.isPipeOut()) {
            os = System.out;
            log.info("Writing results to stdout");
        } else {
            sqlOutFile = new File(dir, fileName);
            os = new BufferedOutputStream(
                    new FileOutputStream(sqlOutFile), 0x10000 /*64k*/);
            log.info("Creating sql file " + sqlOutFile.toString());
        }
        
        os.write((""
            + "-- Created by  : " + Version.getName() + "\n"
            + "-- Version     : " + Version.getVersion() + "\n"
            + "-- Author (c)  : Carsten Moeller - info@osm2po.de\n"
            + "-- Date        : " + new Date().toString()
            + "\n\n"
            + "SET client_encoding = 'UTF8';"
            + "\n\n"
            + "DROP TABLE IF EXISTS " + tableName + ";"
            + "\n"
            + "-- SELECT DropGeometryTable('" + tableName + "');"
            + "\n\n"
            + "CREATE TABLE " + tableName + "("
            + "id integer, osm_id bigint, "
            + "osm_name character varying, osm_meta character varying, "
            + "clazz integer, flags integer, height integer"
            + ");\n"
            + "SELECT AddGeometryColumn('" + tableName + "', "
            + "'geom_poly', 4326, 'POLYGON', 2);\n")
        .getBytes());

        byte[] INSERT = ("\nINSERT INTO " + tableName + " VALUES ").getBytes();
        
        int id = 0, g = 0;
        LatLons latLons = new LatLons();

        for (Way way : new JoinedWaysIterator(config)) {
            byte clazz = way.getClazz();
        
            if (!way.isPolygon()) continue;
            if (!way.isRing()) continue; // Simple Polygon Check
            
            ++id;
            
            long osmId = way.getId();
            int flags = way.getFlags();
            int height = way.getValue();
            byte[] nameSql = toUTF8QuotedSqlBytes(way.getName(), true);
            byte[] metaSql = toUTF8QuotedSqlBytes(way.getMeta(), true);
            
            latLons.setCoords(way.getNodes());
            String geom_poly = bytesToHex(toPolygonWkb(new LatLons[]{latLons}));
            
            // Bulk inserts are shorter and faster
            if (g == 25) {
                os.write(SEMIKOLON);
                os.write(NEWLINE);
                g = 0;
            }
            if (++g == 1) os.write(INSERT); else os.write(KOMMA);
            os.write(NEWLINE);
            
            os.write(("(" + id + ", " + osmId + ", ").getBytes());
            os.write(nameSql);
            os.write(KOMMA);
            os.write(metaSql);
            os.write((", "
                + clazz + ", " + flags + ", " + height + ", "
                + "'" + geom_poly + "'" + ")")
                .getBytes());
            
            if (id % 50000 == 0) log.progress(DF(id) + " Polygons written.");
        }

        os.write(SEMIKOLON);
        os.write(NEWLINE);

        log.info(DF(id) + " Polygons written.");

        os.write(("\n"   
            + "ALTER TABLE " + tableName + " ADD CONSTRAINT pkey_" + tableName + " PRIMARY KEY(id);\n"
            + "-- CREATE INDEX idx_" + tableName + "_geom_poly ON " + tableName
            + " USING GIST (geom_poly);\n")
            .getBytes());

        if (sqlOutFile != null) {
            os.close();
            log.info("commandline template:\n"
                    + "psql -U [username] -d [dbname] -q -f \""
                    + sqlOutFile.getAbsolutePath() + "\"");
        }
        
    }
    
}
